#define _GNU_SOURCE

#include <stdio.h>
#include <xmms/plugin.h>
#include <xmms/configfile.h>
#include <gtk/gtk.h>
#include <math.h>
#include <string.h>

/**************************************************
 * 
 * Speaker Simulator plugin for XMMS
 * 
 * An effect plugin to improve listening with head-phones.
 * Based on the Speaker Simulator Plugin for WinAMP by Vladimir Kopjov
 * It mixes the left channel with some attenuation and delay into the right channel
 * and vice-versa. Makes listening with headphones more "realistic".
 */

/* TODO :
 *  add delay
 * use dithering for better quality
 */

/* Discoveries :
 * when listening to the same sound with both ears, if one is delayed , then it 
 * appears to the listener as if the "earlier" sound is louder (or that the sound
 * source is in the direction of the "earlier" channel )
 * The effect takes most of it strentgh with delays around 200-300 microseconds.
 */

/* Modification by Alexander Thomas, Oct. 2004:
 * Settings are now remembered by storing them in XMMS' preferences file.
 */

#if 0
# define BUG(x) 
#else
#define BUG printf
#endif

#define SPEAKSIM_VERSION "V0.0.3"
#define SAMPLE_BUFFER_BYTES 256  /* buffer size in bytes, enough for 1 millisecond of CD sound , FIXME what if we use extremely large data , like 24 bit 96kHz ?? */
#define DEFAULT_CROSSFADE 0.7
#define DEFAULT_DELAY 250.0

static void init(void);
static void cleanup(void);
static void about(void);
static void configure(void);
static int mod_samples(gpointer *d, gint length, AFormat afmt, gint srate, gint nch);

EffectPlugin stereo_ep =
{
	NULL,
	NULL,
	"Speaker Simulator Plugin " SPEAKSIM_VERSION,
	init,
	cleanup,
	about,
	configure,
	mod_samples,
	NULL
};

static const char about_text[] = "Speaker Simulator Plugin\n\n\
By David Balazic 2000.";

GtkWidget *conf_dialog = NULL, *about_dialog = NULL;
GtkObject *crossfade_adjustment,*delay_adjustment;
gfloat old_crossfade,old_delay;
gint8 sample_buffer[SAMPLE_BUFFER_BYTES];
gint8 sample_buffer2[SAMPLE_BUFFER_BYTES];

EffectPlugin *get_eplugin_info(void)
{
	return &stereo_ep;
}

static void init(void)
{
	/* BEGIN *NEW!* */
	ConfigFile *cfgfile;
	float xmix = DEFAULT_CROSSFADE, delay = DEFAULT_DELAY;
	cfgfile = xmms_cfg_open_default_file();
	xmms_cfg_read_float(cfgfile, "speaksim_plugin", "crossmix", &xmix);
	xmms_cfg_read_float(cfgfile, "speaksim_plugin", "delay", &delay);
	xmms_cfg_free(cfgfile);
	/* END *NEW!* */
	
	/* default crossmix intensity is 70% */
	crossfade_adjustment = gtk_adjustment_new(xmix, 0.0, 1.0 + 0.1, 0.01, 0.1, 0.1); /* *MODIFIED* */
	/* 0.25 millisecs is the default delay */
	delay_adjustment = gtk_adjustment_new(delay, 0.0, 1000.0 + 10.0, 0.1, 1.0, 10.0); /* *MODIFIED* */
	memset(sample_buffer,0,SAMPLE_BUFFER_BYTES);
	memset(sample_buffer2,0,SAMPLE_BUFFER_BYTES);
}

static void cleanup(void)
{
	/* BEGIN *NEW!* */
	ConfigFile *cfgfile;
	cfgfile = xmms_cfg_open_default_file();
	xmms_cfg_write_float(cfgfile, "speaksim_plugin", "crossmix",
						 GTK_ADJUSTMENT(crossfade_adjustment)->value);
	xmms_cfg_write_float(cfgfile, "speaksim_plugin", "delay",
						 GTK_ADJUSTMENT(delay_adjustment)->value);
	xmms_cfg_write_default_file(cfgfile);
	xmms_cfg_free(cfgfile);
	/* END *NEW!* */
	
	gtk_object_unref(crossfade_adjustment);
	gtk_object_unref(delay_adjustment);
}

static gint about_destroy_cb(GtkWidget * w, GdkEventAny * e, gpointer data)
{
	gtk_widget_destroy(about_dialog);
	about_dialog = NULL;
	return TRUE;
}

static void about_ok_cb(GtkButton * button, gpointer data)
{
	gtk_widget_destroy(about_dialog);
	about_dialog = NULL;
}

static void about(void)
{
	GtkWidget *hbox, *text, *button;

	if (about_dialog != NULL)
		return;

	about_dialog = gtk_dialog_new();
	gtk_signal_connect(GTK_OBJECT(about_dialog), "destroy",
			   GTK_SIGNAL_FUNC(about_destroy_cb), NULL);
	gtk_window_set_title(GTK_WINDOW(about_dialog), "Configure Speaker Simulator");

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(about_dialog)->vbox), hbox,
			   TRUE, TRUE, 5);
	gtk_widget_show(hbox);

	text = gtk_text_new(NULL, NULL);
	gtk_text_set_editable(GTK_TEXT(text), FALSE);
	gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL,
			about_text, strlen(about_text));
	gtk_box_pack_start(GTK_BOX(hbox), text, TRUE, TRUE, 5);
	gtk_widget_show(text);

	button = gtk_button_new_with_label("Ok");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(about_dialog)->action_area),
			   button, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
			   GTK_SIGNAL_FUNC(about_ok_cb), NULL);
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_widget_grab_default(button);
	gtk_widget_show(button);
	gtk_widget_show(about_dialog);
}

static gint conf_destroy_cb(GtkWidget * w, GdkEventAny * e, gpointer data)
{
	GTK_ADJUSTMENT(crossfade_adjustment)->value = old_crossfade;
	gtk_signal_emit_by_name(crossfade_adjustment, "changed");
	GTK_ADJUSTMENT(delay_adjustment)->value = old_delay;
	gtk_signal_emit_by_name(delay_adjustment, "changed");
	gtk_widget_destroy(conf_dialog);
	conf_dialog = NULL;
	return TRUE;
}

static void conf_ok_cb(GtkButton * button, gpointer data)
{
	old_crossfade = GTK_ADJUSTMENT(crossfade_adjustment)->value;
	old_delay = GTK_ADJUSTMENT(delay_adjustment)->value;
	gtk_widget_destroy(conf_dialog);
	conf_dialog = NULL;
}

static void conf_cancel_cb(GtkButton * button, gpointer data)
{
	GTK_ADJUSTMENT(crossfade_adjustment)->value = old_crossfade;
	GTK_ADJUSTMENT(delay_adjustment)->value = old_delay;
	gtk_signal_emit_by_name(crossfade_adjustment, "changed");
	gtk_signal_emit_by_name(delay_adjustment, "changed");
	gtk_widget_destroy(conf_dialog);
	conf_dialog = NULL;
}

static void configure(void)
{
	GtkWidget *hbox1,*hbox2;
	GtkWidget *label_crossfade,*label_delay;
	GtkWidget *scale_crossfade;
	GtkWidget *scale_delay;
	GtkWidget *button;


	if (conf_dialog != NULL)
		return;

	if (crossfade_adjustment)
		old_crossfade = GTK_ADJUSTMENT(crossfade_adjustment)->value;
	else
		old_crossfade = DEFAULT_CROSSFADE;
   	if (delay_adjustment)
		old_delay = GTK_ADJUSTMENT(delay_adjustment)->value;
	else
		old_delay = DEFAULT_DELAY;

	conf_dialog = gtk_dialog_new();
	gtk_signal_connect(GTK_OBJECT(conf_dialog), "destroy",
			   GTK_SIGNAL_FUNC(conf_destroy_cb), NULL);
	gtk_window_set_title(GTK_WINDOW(conf_dialog), "Configure Speaker Simulator");

	label_crossfade = gtk_label_new("Crossmix intensity:");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(conf_dialog)->vbox), label_crossfade,
			   TRUE, TRUE, 0);
	gtk_widget_show(label_crossfade);

	hbox1 = gtk_hbox_new(FALSE, 10);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(conf_dialog)->vbox), hbox1,
			   TRUE, TRUE, 10);
	gtk_widget_show(hbox1);

	scale_crossfade = gtk_hscale_new(GTK_ADJUSTMENT(crossfade_adjustment));
	gtk_scale_set_digits((GtkScale *)scale_crossfade,2);
	gtk_object_ref(crossfade_adjustment);
	gtk_box_pack_start(GTK_BOX(hbox1), scale_crossfade, TRUE, TRUE, 10);
	gtk_widget_show(scale_crossfade);

   	label_delay = gtk_label_new("Delay in us:");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(conf_dialog)->vbox), label_delay,
			   TRUE, TRUE, 0);
	gtk_widget_show(label_delay);

	hbox2 = gtk_hbox_new(FALSE, 10);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(conf_dialog)->vbox), hbox2,
			   TRUE, TRUE, 10);
	gtk_widget_show(hbox2);

	scale_delay = gtk_hscale_new(GTK_ADJUSTMENT(delay_adjustment));
	gtk_object_ref(delay_adjustment);
	gtk_box_pack_start(GTK_BOX(hbox2), scale_delay, TRUE, TRUE, 10);
	gtk_widget_show(scale_delay);

	button = gtk_button_new_with_label("Ok");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(conf_dialog)->action_area),
			   button, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
			   GTK_SIGNAL_FUNC(conf_ok_cb), NULL);
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_widget_grab_default(button);
	gtk_widget_show(button);

	button = gtk_button_new_with_label("Cancel");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(conf_dialog)->action_area), button,
			   TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
			   GTK_SIGNAL_FUNC(conf_cancel_cb), NULL);
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_widget_show(button);

	gtk_widget_show(conf_dialog);
}


/* most of 2 is num_channels ! */
#define GET_OLD_LEFT_SAMPLE \
((i-2*delay_offset>=0)?(data[i-2*delay_offset]): \
(((gint16 *)sample_buffer)[SAMPLE_BUFFER_BYTES/2-2*delay_offset+i]))

#define GET_OLD_RIGHT_SAMPLE \
((i-2*delay_offset>=0)?(data[1+i-2*delay_offset]): \
(((gint16 *)sample_buffer)[SAMPLE_BUFFER_BYTES/2-2*delay_offset+i+1]))

#define GET_OLD_LEFT_SAMPLE2 \
((i-2*delay_offset-2>=0)?(data[i-2*delay_offset-2]): \
(((gint16 *)sample_buffer)[SAMPLE_BUFFER_BYTES/2-2*delay_offset+i-2]))

#define GET_OLD_RIGHT_SAMPLE2 \
((i-2*delay_offset-2>=0)?(data[1+i-2*delay_offset-2]): \
(((gint16 *)sample_buffer)[SAMPLE_BUFFER_BYTES/2-2*delay_offset+i+1-2]))

/* FIXME: implement dithering , and other quality improvements */
static int mod_samples(gpointer *d, gint length, AFormat afmt, gint srate, gint nch)
{
	int i,delay_offset;
	double  mul,a,b;
	gint16  *data = (gint16 *)*d;
	gint8   *byte_data = (gint8 *)*d;
	int copy_amount;

	/* handle only 16 bit signed stereo native endian */
	 if (!(afmt == FMT_S16_NE || (afmt == FMT_S16_LE && G_BYTE_ORDER == G_LITTLE_ENDIAN) || (afmt == FMT_S16_BE && G_BYTE_ORDER == G_BIG_ENDIAN)) || nch != 2)
		return length;

	mul = GTK_ADJUSTMENT(crossfade_adjustment)->value;
   /* delay_offset in sample pairs ; 1 means : skip one pair of left-right 16 bit samples ,
    FIXME : fix this for other formats , besides 16bit stereo 44.1 kHz */
	delay_offset=round( (GTK_ADJUSTMENT(delay_adjustment)->value) * 44100.0 /1000000.0);

/*   printf("SS:mod_samples(d=%p;*d=%p, length = %8d...), mul = %8f,delay_offset = %8d = round(%8f)..%d\r",
	  d,*d,length,mul,delay_offset,(GTK_ADJUSTMENT(delay_adjustment)->value) * 44100.0 /1000000.0,101);
*/
   /* move the old data ( try pointer tricks instead of copying , to speed up ) */
   /* nr. of bytes to save */
   copy_amount = (SAMPLE_BUFFER_BYTES>length)?(length):(SAMPLE_BUFFER_BYTES);
   /* save old buffer */
   if (copy_amount < SAMPLE_BUFFER_BYTES )
     memmove(sample_buffer2,sample_buffer+copy_amount,SAMPLE_BUFFER_BYTES-copy_amount);

   /* copy new data */
   memmove(sample_buffer2+SAMPLE_BUFFER_BYTES-copy_amount,byte_data+length-copy_amount,copy_amount);

	for (i = length / 2 - 2 ; i>=0 ; i -= 2) /* go backwards , because we overwrite old data ! */
	{
//		a=(data[i] + mul*GET_OLD_RIGHT_SAMPLE)/(1+mul);
		a=(data[i] + mul*(GET_OLD_RIGHT_SAMPLE+GET_OLD_RIGHT_SAMPLE2)/2)/(1+mul);

		if (a < -32768.0){
			a = -32768.0;
		   printf("CLIP!\n");}
		if (a > 32767.0){
			a = 32767.0;
		   printf("CLIP!\n");}
		data[i] = round(a);

//		b=(data[i+1] + mul*GET_OLD_LEFT_SAMPLE)/(1+mul);
		b=(data[i+1] + mul*(GET_OLD_LEFT_SAMPLE+GET_OLD_LEFT_SAMPLE2)/2)/(1+mul);

		if (b < -32768.0){
			b = -32768.0;
		   printf("CLIP!\n");}
		if (b > 32767.0){
			b = 32767.0;
		   printf("CLIP!\n");}
		data[i + 1] = round(b);
	}

   /* move saved samples */
   memmove(sample_buffer,sample_buffer2,SAMPLE_BUFFER_BYTES);

	return length;
}
